"""Oauth helpers for YouTube."""

from __future__ import annotations

import aiohttp

from youtubeaio.helper import YOUTUBE_AUTH_TOKEN_URL, build_url
from youtubeaio.types import InvalidRefreshTokenError, UnauthorizedError

__all__ = ["refresh_access_token"]


async def refresh_access_token(
    refresh_token: str,
    app_id: str,
    app_secret: str,
    session: aiohttp.ClientSession | None = None,
) -> tuple[str, str]:
    """Refresh a user access token.

    :param str refresh_token: the current refresh_token
    :param str app_id: the id of your app
    :param str app_secret: the secret key of your app
    :param ~aiohttp.ClientSession session: optionally a active client session to be
    used for the web request to avoid having to open a new one
    :return: access_token, refresh_token
    :raises ~youtubeaio.types.InvalidRefreshTokenException: if refresh token
    is invalid
    :raises ~youtubeaio.types.UnauthorizedException: if both refresh and
    access token are invalid (e.g. if the user changes their password of the app gets
    disconnected)
    :rtype: (str, str)
    """
    param = {
        "refresh_token": refresh_token,
        "client_id": app_id,
        "grant_type": "refresh_token",
        "client_secret": app_secret,
        "access_type": "offline",
        "prompt": "consent",
    }
    url = build_url(YOUTUBE_AUTH_TOKEN_URL, {})
    ses = session if session is not None else aiohttp.ClientSession()
    async with ses.post(url, data=param) as result:
        data = await result.json()
    if session is None:
        await ses.close()
    if result.status == 400:
        raise InvalidRefreshTokenError(data.get("error", ""))
    if result.status == 401:
        raise UnauthorizedError(data.get("error", ""))
    return data["access_token"], refresh_token



"""Models for YouTube API."""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import TypeVar

from pydantic import BaseModel, Field

from youtubeaio.const import (
    LiveBroadcastContent,
    VideoDefinition,
    VideoDimension,
    VideoProjection,
)

__all__ = [
    "YouTubeChannel",
    "YouTubeChannelContentDetails",
    "YouTubeChannelRelatedPlaylists",
    "YouTubeChannelSnippet",
    "YouTubeChannelThumbnails",
    "YouTubeThumbnail",
    "YouTubeVideo",
    "YouTubeVideoSnippet",
    "YouTubeVideoThumbnails",
]

from youtubeaio.helper import get_duration
from youtubeaio.types import PartMissingError

T = TypeVar("T")


class YouTubeThumbnail(BaseModel):
    """Model representing a video thumbnail."""

    url: str = Field(...)
    width: int = Field(...)
    height: int = Field(...)


class YouTubeVideoThumbnails(BaseModel):
    """Model representing video thumbnails."""

    default: YouTubeThumbnail = Field(...)
    medium: YouTubeThumbnail | None = Field(None)
    high: YouTubeThumbnail | None = Field(None)
    standard: YouTubeThumbnail | None = Field(None)
    maxres: YouTubeThumbnail | None = Field(None)

    def get_highest_quality(self) -> YouTubeThumbnail:
        """Return the highest quality thumbnail."""
        for size in (self.maxres, self.standard, self.high, self.medium):
            if size is not None:
                return size
        return self.default


class YouTubeVideoSnippet(BaseModel):
    """Model representing video snippet."""

    published_at: datetime = Field(..., alias="publishedAt")
    channel_id: str = Field(..., alias="channelId")
    title: str = Field(...)
    description: str = Field(...)
    thumbnails: YouTubeVideoThumbnails = Field(...)
    channel_title: str = Field(..., alias="channelTitle")
    tags: list[str] = Field([])
    live_broadcast_content: LiveBroadcastContent = Field(
        ...,
        alias="liveBroadcastContent",
    )
    default_language: str | None = Field(None, alias="defaultLanguage")
    default_audio_language: str | None = Field(None, alias="defaultAudioLanguage")


class YouTubeVideoContentDetails(BaseModel):
    """Model representing video content details."""

    raw_duration: str = Field(..., alias="duration")
    dimension: VideoDimension = Field(...)
    definition: VideoDefinition = Field(...)
    raw_caption: str = Field(..., alias="caption")
    licensed_content: bool = Field(..., alias="licensedContent")
    projection: VideoProjection = Field(...)

    @property
    def caption(self) -> bool:
        """Return if video has caption."""
        return self.raw_caption == "true"

    @property
    def duration(self) -> timedelta:
        """Return length of the video."""
        return get_duration(self.raw_duration)


class YouTubeVideo(BaseModel):
    """Model representing a video."""

    video_id: str = Field(..., alias="id")
    nullable_snippet: YouTubeVideoSnippet | None = Field(None, alias="snippet")
    nullable_content_details: YouTubeVideoContentDetails | None = Field(
        None,
        alias="contentDetails",
    )

    @property
    def snippet(self) -> YouTubeVideoSnippet:
        """Return snippet."""
        if self.nullable_snippet is None:
            raise PartMissingError
        return self.nullable_snippet

    @property
    def content_details(self) -> YouTubeVideoContentDetails:
        """Return content details."""
        if self.nullable_content_details is None:
            raise PartMissingError
        return self.nullable_content_details


class YouTubeChannelThumbnails(BaseModel):
    """Model representing channel thumbnails."""

    default: YouTubeThumbnail = Field(...)
    medium: YouTubeThumbnail | None = Field(None)
    high: YouTubeThumbnail | None = Field(None)

    def get_highest_quality(self) -> YouTubeThumbnail:
        """Return the highest quality thumbnail."""
        for size in (self.high, self.medium):
            if size is not None:
                return size
        return self.default


class YouTubeChannelRelatedPlaylists(BaseModel):
    """Model representing related playlists of a channel."""

    likes: str = Field(...)
    uploads: str = Field(...)


class YouTubeChannelContentDetails(BaseModel):
    """Model representing content details of a channel."""

    related_playlists: YouTubeChannelRelatedPlaylists = Field(
        ...,
        alias="relatedPlaylists",
    )


class YouTubeChannelStatistics(BaseModel):
    """Model representing statistics of a channel."""

    view_count: int = Field(..., alias="viewCount")
    subscriber_count: int = Field(..., alias="subscriberCount")
    subscriber_count_hidden: bool = Field(..., alias="hiddenSubscriberCount")
    video_count: int = Field(..., alias="videoCount")


class YouTubeChannelSnippet(BaseModel):
    """Model representing channel snippet."""

    title: str = Field(...)
    description: str = Field(...)
    published_at: datetime = Field(..., alias="publishedAt")
    thumbnails: YouTubeChannelThumbnails = Field(...)
    default_language: str | None = Field(None, alias="defaultLanguage")


class YouTubeChannel(BaseModel):
    """Model representing a YouTube channel."""

    channel_id: str = Field(..., alias="id")
    nullable_snippet: YouTubeChannelSnippet | None = Field(None, alias="snippet")
    nullable_content_details: YouTubeChannelContentDetails | None = Field(
        None,
        alias="contentDetails",
    )
    nullable_statistics: YouTubeChannelStatistics | None = Field(
        None,
        alias="statistics",
    )

    @property
    def upload_playlist_id(self) -> str:
        """Return playlist id with uploads from channel."""
        return str(self.channel_id).replace("UC", "UU", 1)

    @property
    def snippet(self) -> YouTubeChannelSnippet:
        """Return snippet."""
        if self.nullable_snippet is None:
            raise PartMissingError
        return self.nullable_snippet

    @property
    def content_details(self) -> YouTubeChannelContentDetails:
        """Return content details."""
        if self.nullable_content_details is None:
            raise PartMissingError
        return self.nullable_content_details

    @property
    def statistics(self) -> YouTubeChannelStatistics:
        """Return statistics."""
        if self.nullable_statistics is None:
            raise PartMissingError
        return self.nullable_statistics


class YouTubeSubscriptionSnippet(BaseModel):
    """Model representing a YouTube subscription snippet."""

    title: str = Field(...)
    description: str = Field(...)
    subscribed_at: datetime = Field(..., alias="publishedAt")
    channel_info: dict[str, str] = Field(..., alias="resourceId")

    @property
    def channel_id(self) -> str:
        """Return channel id."""
        return self.channel_info["channelId"]


class YouTubeSubscription(BaseModel):
    """Model representing a YouTube subscription."""

    subscription_id: str = Field(..., alias="id")
    nullable_snippet: YouTubeSubscriptionSnippet | None = Field(None, alias="snippet")

    @property
    def snippet(self) -> YouTubeSubscriptionSnippet:
        """Return snippet."""
        if self.nullable_snippet is None:
            raise PartMissingError
        return self.nullable_snippet


class YouTubePlaylistItemSnippet(BaseModel):
    """Model representing a YouTube playlist item snippet."""

    added_at: datetime = Field(..., alias="publishedAt")
    title: str = Field(...)
    description: str = Field(...)
    thumbnails: YouTubeVideoThumbnails = Field(...)
    playlist_id: str = Field(..., alias="playlistId")


class YouTubePlaylistItemContentDetails(BaseModel):
    """Model representing a YouTube playlist item content details."""

    video_id: str = Field(..., alias="videoId")


class YouTubePlaylistItem(BaseModel):
    """Model representing a YouTube playlist item."""

    playlist_item_id: str = Field(..., alias="id")
    nullable_snippet: YouTubePlaylistItemSnippet | None = Field(None, alias="snippet")
    nullable_content_details: YouTubePlaylistItemContentDetails | None = Field(
        None,
        alias="contentDetails",
    )

    @property
    def snippet(self) -> YouTubePlaylistItemSnippet:
        """Return snippet."""
        if self.nullable_snippet is None:
            raise PartMissingError
        return self.nullable_snippet

    @property
    def content_details(self) -> YouTubePlaylistItemContentDetails:
        """Return content details."""
        if self.nullable_content_details is None:
            raise PartMissingError
        return self.nullable_content_details


YouTubeChannelSnippet.model_rebuild()



"""Type definitions."""

from enum import Enum

__all__ = [
    "AuthScope",
    "DeprecatedError",
    "ForbiddenError",
    "InvalidRefreshTokenError",
    "InvalidTokenError",
    "MissingAppSecretError",
    "MissingScopeError",
    "UnauthorizedError",
    "YouTubeAPIError",
    "YouTubeAuthorizationError",
    "YouTubeBackendError",
    "YouTubeResourceNotFoundError",
]


class AuthScope(Enum):
    """Enum of authentication scopes."""

    MANAGE = "https://www.googleapis.com/auth/youtube"
    MANAGE_MEMBERSHIPS = (
        "https://www.googleapis.com/auth/youtube.channel-memberships.creator"
    )
    FORCE_SSL = "https://www.googleapis.com/auth/youtube.force-ssl"
    READ_ONLY = "https://www.googleapis.com/auth/youtube.readonly"
    UPLOAD = "https://www.googleapis.com/auth/youtube.upload"
    PARTNER = "https://www.googleapis.com/auth/youtubepartner"
    PARTNER_AUDIT = "https://www.googleapis.com/auth/youtubepartner-channel-audit"


class YouTubeAPIError(Exception):
    """Base YouTube API exception."""


class YouTubeAuthorizationError(YouTubeAPIError):
    """Exception in the YouTube Authorization."""


class InvalidRefreshTokenError(YouTubeAPIError):
    """used User Refresh Token is invalid."""


class InvalidTokenError(YouTubeAPIError):
    """Used if an invalid token is set for the client."""


class UnauthorizedError(YouTubeAuthorizationError):
    """Not authorized to use this."""


class MissingScopeError(YouTubeAuthorizationError):
    """authorization is missing scope."""


class YouTubeBackendError(YouTubeAPIError):
    """When the YouTube API itself is down."""


class PartMissingError(YouTubeAPIError):
    """If you request a part which is not requested."""


class MissingAppSecretError(YouTubeAPIError):
    """When the app secret is not set but app authorization is attempted."""


class DeprecatedError(YouTubeAPIError):
    """If something has been marked as deprecated by the YouTube API."""


class YouTubeResourceNotFoundError(YouTubeAPIError):
    """If a requested resource was not found."""


class ForbiddenError(YouTubeAPIError):
    """If you are not allowed to do that."""



"""Helper functions for the YouTube API."""

from __future__ import annotations

import re
import urllib.parse
from datetime import timedelta
from enum import Enum
from typing import TYPE_CHECKING, Any, TypeVar

from youtubeaio.types import AuthScope

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Generator

__all__ = [
    "YOUTUBE_AUTH_BASE_URL",
    "YOUTUBE_AUTH_TOKEN_URL",
    "build_scope",
    "build_url",
    "chunk",
    "first",
    "limit",
]

T = TypeVar("T")

YOUTUBE_AUTH_BASE_URL: str = "https://oauth2.googleapis.com"
YOUTUBE_AUTH_TOKEN_URL: str = f"{YOUTUBE_AUTH_BASE_URL}/token"


def build_scope(scopes: list[AuthScope]) -> str:
    """Build a valid scope string from list.

    :param scopes: list of :class:`~youtubeaio.types.AuthScope`
    :returns: the valid auth scope string
    """
    return " ".join([s.value for s in scopes])


def build_url(
    url: str,
    params: dict[str, Any],
    remove_none: bool = False,
    split_lists: bool = False,
    enum_value: bool = True,
) -> str:
    """Build a valid url string.

    :param url: base URL
    :param params: dictionary of URL parameter
    :param remove_none: if set all params that have a None value get removed
    |default| :code:`False`
    :param split_lists: if set all params that are a list will be split over multiple
    url parameter with the same name |default| :code:`False`
    :param enum_value: if true, automatically get value string from Enum values
    |default| :code:`True`
    :return: URL
    """

    def get_value(input_value: Any) -> str:
        if not enum_value:
            return str(input_value)
        if isinstance(input_value, Enum):
            return str(input_value.value)
        return str(input_value)

    def add_param(res: str, query_key: str, query_value: Any) -> str:
        if len(res) > 0:
            res += "&"
        res += query_key
        if query_value is not None:
            res += "=" + urllib.parse.quote(get_value(query_value))
        return res

    result = ""
    for key, value in params.items():
        if value is None and remove_none:
            continue
        if split_lists and isinstance(value, list):
            for val in value:
                result = add_param(result, key, val)
        else:
            result = add_param(result, key, value)
    return url + (("?" + result) if len(result) > 0 else "")


async def first(generator: AsyncGenerator[T, None]) -> T | None:
    """Return the first value or None from the given AsyncGenerator."""
    try:
        return await generator.__anext__()
    except StopAsyncIteration:
        return None


def chunk(source: list[T], chunk_size: int) -> Generator[list[T], None, None]:
    """Divide the source list in chunks of given size."""
    for i in range(0, len(source), chunk_size):
        yield source[i : i + chunk_size]


async def limit(
    generator: AsyncGenerator[T, None],
    total: int,
) -> AsyncGenerator[T, None]:
    """Limit the number of entries returned from the AsyncGenerator."""
    if total < 1:
        msg = "Limit has to be an int > 1"
        raise ValueError(msg)
    count = 0
    async for item in generator:
        count += 1
        if count > total:
            break
        yield item


def get_duration(duration: str) -> timedelta:
    """Return timedelta for ISO8601 duration string."""
    attributes = {
        "S": 0,
        "M": 0,
        "H": 0,
        "D": 0,
    }
    for match in re.compile(r"(\d+[DHMS])").finditer(duration):
        part = match.group(1)
        time_value = int(part[:-1])
        attributes[part[len(part) - 1]] = time_value
    return timedelta(
        days=attributes["D"],
        hours=attributes["H"],
        minutes=attributes["M"],
        seconds=attributes["S"],
    )



"""The YouTube API."""

from __future__ import annotations

import asyncio
from logging import getLogger
from typing import TYPE_CHECKING, Any, TypeVar

from aiohttp import ClientError, ClientResponse, ClientSession

from youtubeaio.helper import (
    build_url,
    first,
)
from youtubeaio.models import (
    YouTubeChannel,
    YouTubePlaylistItem,
    YouTubeSubscription,
    YouTubeVideo,
)
from youtubeaio.types import (
    AuthScope,
    ForbiddenError,
    MissingScopeError,
    UnauthorizedError,
    YouTubeAPIError,
    YouTubeBackendError,
    YouTubeResourceNotFoundError,
)

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable, Coroutine

    from typing_extensions import Self

__all__ = [
    "YouTube",
]
T = TypeVar("T")


class YouTube:
    """YouTube API client."""

    base_url: str = "https://youtube.googleapis.com/youtube/v3/"
    _close_session: bool = False

    logger = getLogger(__name__)

    _user_auth_token: str | None = None
    _user_auth_refresh_token: str | None = None
    _has_user_auth = False

    def __init__(
        self,
        app_id: str | None = None,
        app_secret: str | None = None,
        session: ClientSession | None = None,
        session_timeout: int = 10,
        auto_refresh_auth: bool | None = None,
    ) -> None:
        """Initialize YouTube object."""
        self.session = session
        self.session_timeout = session_timeout
        self.app_id = app_id
        self.app_secret = app_secret
        self._user_auth_scopes: list[AuthScope] = []
        if auto_refresh_auth is None:
            self.auto_refresh_auth = app_id is not None and app_secret is not None
        else:
            self.auto_refresh_auth = auto_refresh_auth
        self._r_lookup: dict[
            str,
            Callable[
                [ClientSession, str, dict[str, Any] | None],
                Coroutine[Any, Any, ClientResponse],
            ],
        ] = {
            "get": self._api_get_request,
            "head": self._api_head_request,
        }

    async def _check_request_return(self, response: ClientResponse) -> ClientResponse:
        if response.status == 500:
            msg = "Internal Server Error"
            raise YouTubeBackendError(msg)
        if response.status == 400:
            msg = (await response.json()).get("message")
            raise YouTubeAPIError(
                "Bad Request" + ("" if msg is None else f" - {msg!s}"),
            )
        if response.status == 404:
            raise YouTubeResourceNotFoundError
        if response.status == 401:
            raise UnauthorizedError
        if response.status == 403:
            response_json = await response.json()
            error_message = response_json["error"]["errors"][0]["message"]
            raise ForbiddenError(error_message)
        if 400 <= response.status < 500:
            try:
                response.raise_for_status()
            except ClientError as exc:
                raise YouTubeAPIError from exc
        return response

    async def _api_get_request(
        self,
        session: ClientSession,
        url: str,
        data: dict[str, Any] | None = None,
    ) -> ClientResponse:
        """Make GET request with authorization."""
        headers = {"Authorization": f"Bearer {self._user_auth_token}"}
        self.logger.debug("making GET request to %s", url)
        response = await session.get(url, headers=headers, json=data)
        return await self._check_request_return(response)

    async def _api_head_request(
        self,
        session: ClientSession,
        url: str,
        data: dict[str, Any] | None = None,
    ) -> ClientResponse:
        """Make HEAD request with authorization."""
        headers = {"Authorization": f"Bearer {self._user_auth_token}"}
        self.logger.debug("making HEAD request to %s", url)
        response = await session.head(url, headers=headers, json=data)
        return await self._check_request_return(response)

    async def _build_generator(
        self,
        req: str,
        url: str,
        url_params: dict[str, Any],
        return_type: T,
        body_data: dict[str, Any] | None = None,
        split_lists: bool = False,
    ) -> AsyncGenerator[T, None]:
        method = self._r_lookup.get(req.lower(), self._api_get_request)
        _after = url_params.get("nextPageToken")
        _first = True
        if not self.session:
            self.session = ClientSession()
            self._close_session = True
        try:
            while _first or _after is not None:
                url_params["pageToken"] = _after
                _url = build_url(
                    self.base_url + url,
                    url_params,
                    remove_none=True,
                    split_lists=split_lists,
                )
                async with asyncio.timeout(self.session_timeout):
                    response = await method(self.session, _url, body_data)
                if response.content_type != "application/json":
                    msg = "Unexpected response type"
                    raise YouTubeAPIError(msg)
                data = await response.json()
                for entry in data.get("items", []):
                    yield return_type(**entry)  # type: ignore[operator]
                _after = data.get("nextPageToken")
                _first = False
        except asyncio.TimeoutError as exc:
            msg = "Timeout occurred"
            raise YouTubeBackendError(msg) from exc

    async def set_user_authentication(
        self,
        token: str,
        scopes: list[AuthScope],
        refresh_token: str | None = None,
    ) -> None:
        """Set a user token to be used.

        :param token: the generated user token
        :param scopes: List of Authorization Scopes that the given user token has
        :param refresh_token: The generated refresh token, has to be provided if
        :attr:`auto_refresh_auth` is True |default| :code:`None`
        :raises ValueError: if :attr:`auto_refresh_auth` is True but refresh_token is
        not set
        :raises ~youtubeaio.types.MissingScopeException: if given token is
         missing one of the required scopes
        :raises ~youtubeaio.types.InvalidTokenException: if the given token
        is invalid or for a different client id
        """
        if refresh_token is None and self.auto_refresh_auth:
            msg = "refresh_token has to be provided when auto_refresh_auth is True"
            raise ValueError(msg)
        if not scopes:
            msg = "scope was not provided"
            raise MissingScopeError(msg)

        self._user_auth_token = token
        self._user_auth_refresh_token = refresh_token
        self._user_auth_scopes = scopes
        self._has_user_auth = True

    def get_user_auth_token(self) -> str | None:
        """Return the current user auth token, None if no user Authentication is set.

        :return: current user auth token
        """
        return self._user_auth_token

    async def get_videos(
        self,
        video_ids: list[str],
    ) -> AsyncGenerator[YouTubeVideo, None]:
        """Get videos by id."""
        if not video_ids:
            msg = "at least one video id has to be set"
            raise ValueError(msg)
        param = {
            "part": "snippet",
            "id": ",".join(video_ids),
        }
        async for item in self._build_generator(
            "GET",
            "videos",
            param,
            YouTubeVideo,
        ):
            yield item  # type: ignore[misc]

    async def get_video(self, video_id: str) -> YouTubeVideo | None:
        """Get a single video."""
        return await first(self.get_videos([video_id]))

    async def _get_channels(
        self,
        param: dict[str, Any],
    ) -> AsyncGenerator[YouTubeChannel, None]:
        """Get channels."""
        async for item in self._build_generator(
            "GET",
            "channels",
            param,
            YouTubeChannel,
        ):
            yield item  # type: ignore[misc]

    async def get_user_channels(self) -> AsyncGenerator[YouTubeChannel, None]:
        """Return channels owned by the authenticated user."""
        param = {
            "part": "snippet",
            "mine": "true",
        }
        async for item in self._get_channels(param):
            yield item

    async def get_channels(
        self,
        channel_ids: list[str],
    ) -> AsyncGenerator[YouTubeChannel, None]:
        """Return list of channels."""
        param = {
            "part": "snippet,contentDetails,statistics",
            "id": ",".join(channel_ids),
        }
        async for item in self._get_channels(param):
            yield item

    async def get_user_subscriptions(
        self,
    ) -> AsyncGenerator[YouTubeSubscription, None]:
        """Get subscriptions for authenticated user."""
        param = {
            "part": "snippet",
            "mine": "true",
        }
        async for item in self._build_generator(
            "GET",
            "subscriptions",
            param,
            YouTubeSubscription,
        ):
            yield item  # type: ignore[misc]

    async def get_playlist_items(
        self,
        playlist_id: str,
        max_results: int = 50,
    ) -> AsyncGenerator[YouTubePlaylistItem, None]:
        """Get playlist by id."""
        param = {
            "part": "snippet,contentDetails",
            "playlistId": playlist_id,
            "maxResults": max_results,
        }
        async for item in self._build_generator(
            "GET",
            "playlistItems",
            param,
            YouTubePlaylistItem,
        ):
            yield item  # type: ignore[misc]

    async def is_short(self, video_id: str) -> bool:
        """Return True if the video ID corresponds to a YouTube Short."""
        _url = "https://www.youtube.com/shorts/" + video_id
        if not self.session:
            self.session = ClientSession()
            self._close_session = True
        async with asyncio.timeout(self.session_timeout):
            response = await self._api_head_request(self.session, _url)
        if response.status == 200:
            return True
        if response.status == 303:
            return False
        raise YouTubeAPIError

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The YouTube object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self.close()
